#include "stdafx.h"

#define STR_OS_LIST_SECTION  "[Operating Systems]"
#define STR_BOOTLDR_SECTION  "[Boot Loader]"
#define REG_PATH_CONTROL "SYSTEM\\CurrentControlSet\\Control"
#define REG_VAL_SYS_BOOT_DRV "SystemBootDevice"

HANDLE StdIn;
HANDLE StdOut;
BOOLEAN OwnConsole = FALSE;

UCHAR BootDrv[64] = "";
    UCHAR BootIniPath[] = "x:\\boot.ini";

//extern "C"
VOID
printf(
  PCHAR DebugMessage,
  ...
  )
{

    CHAR dbg_print_tmp_buff[2048];
    ULONG len;

    va_list ap;
    va_start(ap, DebugMessage);

    len = _vsnprintf(&dbg_print_tmp_buff[0], 2047, DebugMessage, ap);

    dbg_print_tmp_buff[2047] = 0;

    WriteConsole(StdOut, &dbg_print_tmp_buff[0], len, &len, NULL);

    va_end(ap);
}

typedef struct _BOOT_INI_LINE_DESCR {
    ULONG LineN;
    ULONG dupLineN;
    ULONG len;
    PUCHAR begin;
    PUCHAR end;
    ULONG  dev_len;
    PUCHAR dev_begin;
    PUCHAR dev_end;
    ULONG  title_len;
    PUCHAR title_begin;
    PUCHAR title_end;
    BOOLEAN boot_item;
    BOOLEAN def_item;
    BOOLEAN orig_def_item;
} BOOT_INI_LINE_DESCR, *PBOOT_INI_LINE_DESCR;

ULONG
drv_letter_to_index(
    WCHAR a
    )
{
    if(a >= 'a' && a <= 'z') {
        return a - 'a';
    }
    if(a >= 'A' && a <= 'Z') {
        return a - 'A';
    }
    return -1;
} // end drv_letter_to_index()

ULONG
num_or_letter_to_index(
    WCHAR a
    )
{
    if(a >= 'a' && a <= 'z') {
        return a - 'a' + 9;
    }
    if(a >= 'A' && a <= 'Z') {
        return a - 'A' + 9;
    }
    if(a >= '1' && a <= '9') {
        return a - '1';
    }
    return -1;
} // end num_or_letter_to_index()

int skip_spaces(
    PUCHAR lpCmdLine
    )
{
    ULONG i;
    BOOLEAN break_name = FALSE;
    for (i=0; !break_name ;i++) {
        break_name = TRUE;
        switch(lpCmdLine[i]) {
        case 0:
        case ' ':
        case '\t':
            break_name = FALSE;
        }
        if(break_name)
            break;
    }
    return i;
} // end skip_spaces()

void
set_attr(
    USHORT attr
    )
{
//    fflush(stdout);
    SetConsoleTextAttribute(StdOut, attr);
} // end set_attr()

int
get_qname(
    PUCHAR lpCmdLine
    )
{
    BOOLEAN break_on_q = FALSE;
    BOOLEAN break_name = FALSE;
    ULONG i;

    // skip name of executable
    if(lpCmdLine[0] == '\"') {
        // check if executable name comes in quotation marks
        break_on_q = TRUE;
    }
    // search for end of executable's name
    for (i=break_on_q; !break_name /*&& (i<lim-1)*/ ;i++) {
        switch(lpCmdLine[i]) {
        case '\"':
            break_name = TRUE;
            break;
        case 0:
        case '\n':
        case '\r':
            break_name = TRUE;
            break;
        case ' ':
        case '=':
        case '\t':
//        case '\n':
//        case '\r':
            if(!break_on_q) {
                break_name = TRUE;
            }
            break;
        }
    }
    return i-1;
} // end get_qname()

LONGLONG
get_writetime_by_name(
    char* name
    )
{
    HANDLE h;
    LONGLONG ftime;

    h = CreateFile(name, GENERIC_READ, FILE_SHARE_READ, NULL,
                      OPEN_EXISTING,
                      FILE_ATTRIBUTE_NORMAL,
                      NULL);
    if(h == (HANDLE)(-1)) {
        return 0;
    }
    if(!GetFileTime(h, NULL, NULL, (PFILETIME)&ftime)) {
        ftime = 0;
    }
    CloseHandle(h);
    return ftime;
} // end get_writetime_by_name()

PBOOT_INI_LINE_DESCR
parse_boot_ini(
    char* BootIniPath
    )
{
    HANDLE h;
    ULONG sz;
    PUCHAR buff;
    ULONG read_bytes;
    ULONG i, n, m, d, cnt;
    BOOLEAN found;
    UCHAR a;
    ULONG lc;
    ULONG line_ldr=-1;
    ULONG line_lst=-1;
    ULONG section = 0; // 1=ldr, 2=lst
    CHAR BootIniBkpPath[] = "x:\\boot.prv";

    BOOLEAN new_line;
    BOOLEAN in_qm = FALSE;
    PUCHAR boot_path = NULL;
    BOOLEAN wait_boot_path = FALSE;
    PUCHAR boot_name = NULL;
    BOOLEAN wait_boot_name = FALSE;
    BOOLEAN use_copy;

    PBOOT_INI_LINE_DESCR llist;

    BootIniBkpPath[0] = BootIniPath[0];

    if(get_writetime_by_name(BootIniPath) > get_writetime_by_name(BootIniBkpPath)) {
        use_copy = FALSE;
    } else {
        use_copy = TRUE;
    }

    h = CreateFile(use_copy ? BootIniBkpPath : BootIniPath,
                      GENERIC_READ, FILE_SHARE_READ, NULL,
                      OPEN_EXISTING,
                      FILE_ATTRIBUTE_NORMAL,
                      NULL);
    if(h == (HANDLE)(-1)) {
        printf("Can't open %s\n", BootIniPath);
        return NULL;
    }
    sz = GetFileSize(h, NULL);
    buff = (PUCHAR)GlobalAlloc(GMEM_FIXED, sz+1);
    if(!buff) {
        printf("Insufficient memory: can't allocate %d bytes\n", sz);
        return NULL;
    }
    if(!ReadFile(h, buff, sz, &read_bytes, NULL) ||
       read_bytes != sz) {
        CloseHandle(h);
        GlobalFree(buff);
        printf("Can't read %s\n", BootIniPath);
        return NULL;
    }
    CloseHandle(h);
    buff[sz] = 0;

    if(!use_copy) {
        SetFileAttributes((PCHAR)BootIniBkpPath, FILE_ATTRIBUTE_NORMAL);
        h = CreateFile((PCHAR)BootIniBkpPath, GENERIC_WRITE, FILE_SHARE_READ, NULL,
                          CREATE_ALWAYS,
                          FILE_ATTRIBUTE_NORMAL,
                          NULL);
        if(h == (HANDLE)(-1)) {
            printf("Can't make backup copy %s\n", BootIniBkpPath);
            return NULL;
        }
        if(!WriteFile(h, buff, sz, &read_bytes, NULL) ||
           read_bytes != sz) {
            CloseHandle(h);
            GlobalFree(buff);
            printf("Can't write %s\n", BootIniBkpPath);
            return NULL;
        }
        CloseHandle(h);
        SetFileAttributes((PCHAR)BootIniBkpPath, FILE_ATTRIBUTE_READONLY | FILE_ATTRIBUTE_SYSTEM | FILE_ATTRIBUTE_HIDDEN);
    }

    // calculate lines
    new_line = TRUE;
    lc = 1;
    for(i=0; i<sz; i++) {
        a = buff[i];
        if(a == '\n' || a == '\r') {
            if(new_line) {
                lc++;
                new_line = FALSE;
            }
        } else {
            if(!new_line) {
                new_line = TRUE;
            }
        }
    }

    llist = (PBOOT_INI_LINE_DESCR)GlobalAlloc(GMEM_FIXED, (lc+1)*sizeof(BOOT_INI_LINE_DESCR));
    memset(llist, 0, (lc+1)*sizeof(BOOT_INI_LINE_DESCR));
    llist[0].begin = buff;
    llist[0].LineN = 0;
    lc = 0;
    new_line = TRUE;
    for(i=0; i<sz; i++) {
        a = buff[i];
        if(a == '\n' || a == '\r') {
            if(new_line) {
                lc++;
                new_line = FALSE;
            }
        } else {
            if(!new_line) {
                llist[lc-1].end = buff+i;
                llist[lc-1].len = (ULONG)(llist[lc-1].end) - (ULONG)(llist[lc-1].begin);
                llist[lc].begin = buff+i;
                llist[lc].LineN = lc;
                new_line = TRUE;
            }
        }
    }
    llist[lc-1].end = buff+i;
    llist[lc-1].len = (ULONG)(llist[lc-1].end) - (ULONG)(llist[lc-1].begin);

    found = FALSE;
    for(n=0; n<lc && !found; n++) {
        // locate [Boot Loader] section
        if(llist[n].len < sizeof(STR_BOOTLDR_SECTION) - sizeof(CHAR))
            continue;
        for(i=0; i<llist[n].len-sizeof(STR_BOOTLDR_SECTION); i++) {
            if(llist[n].begin[i] == ';')
                break;
            if(!_strnicmp((CCHAR*)(llist[n].begin+i), STR_BOOTLDR_SECTION, sizeof(STR_BOOTLDR_SECTION)-sizeof(CHAR))) {
                found = TRUE;
                line_ldr=n;
                break;
            }
        }
    }
    if(!found) {
        printf("Can't find " STR_BOOTLDR_SECTION " section\n");
        GlobalFree(buff);
        return NULL;
    }

    found = FALSE;
    for(n=0; n<lc && !found; n++) {
        // locate [Operating Systems] section
        if(llist[n].len < sizeof(STR_OS_LIST_SECTION) - sizeof(CHAR))
            continue;
        for(i=0; i<llist[n].len-sizeof(STR_OS_LIST_SECTION); i++) {
            if(llist[n].begin[i] == ';')
                break;
            if(!_strnicmp((CCHAR*)(llist[n].begin+i), STR_OS_LIST_SECTION, sizeof(STR_OS_LIST_SECTION)-sizeof(CHAR))) {
                found = TRUE;
                line_lst=n;
                break;
            }
        }
    }
    if(!found) {
        printf("Can't find " STR_OS_LIST_SECTION " section\n");
        GlobalFree(buff);
        return NULL;
    }

    for(n=0; n<lc; n++) {
        if(n == line_ldr) {
            section=1;
        } else
        if(n == line_lst) {
            section=2;
        } else
        { /* do nothing */; }

        cnt = 0;
        d = skip_spaces(llist[n].begin);
        cnt += d;
        if(cnt >= llist[n].len)
            continue;
        llist[n].dev_begin = llist[n].begin+d;
        // read ArcName
        d = get_qname(llist[n].dev_begin);
        cnt += d;
        if(cnt >= llist[n].len) {
            llist[n].dev_begin = NULL;
            continue;
        }
        llist[n].dev_end = llist[n].dev_begin+d;
        llist[n].dev_len = (ULONG)(llist[n].dev_end) - (ULONG)(llist[n].dev_begin);
        // locate '=' sign
        d = skip_spaces(llist[n].dev_end);
        cnt += d;
        if(cnt >= llist[n].len) {
            llist[n].dev_begin = NULL;
            llist[n].dev_end = NULL;
            continue;
        }
        for(i=cnt; i<llist[n].len; i++) {
            if(llist[n].begin[i] == '=')
                break;
        }
        if(llist[n].begin[i] != '=') {
            llist[n].dev_begin = NULL;
            llist[n].dev_end = NULL;
            continue;
        }
        // read Title
        cnt = i+1;
        d = skip_spaces(llist[n].begin+cnt);
        cnt += d;
        if(cnt >= llist[n].len) {
            llist[n].dev_begin = NULL;
            llist[n].dev_end = NULL;
            continue;
        }
        llist[n].title_begin = llist[n].begin+cnt;
        d = get_qname(llist[n].title_begin);
        cnt += d;
        if(cnt >= llist[n].len) {
            llist[n].dev_begin = NULL;
            llist[n].dev_end = NULL;
            llist[n].title_begin = NULL;
            continue;
        }
        llist[n].title_end = llist[n].title_begin+d;
        if(llist[n].title_begin[0] == '\"') {
            llist[n].title_begin++;
        }
        llist[n].title_len = (ULONG)(llist[n].title_end) - (ULONG)(llist[n].title_begin);
        if(section == 1) {
            if(!_strnicmp((CCHAR*)(llist[n].dev_begin), "Default", sizeof("Default")-sizeof(CHAR))) {
                llist[n].def_item = TRUE;
            } else
            if(!_strnicmp((CCHAR*)(llist[n].dev_begin), ";OrigDefault", sizeof(";OrigDefault")-sizeof(CHAR))) {
                llist[n].orig_def_item = TRUE;
            } else
            if(llist[n].dev_begin[0] == ';') {
                // skip comment
                llist[n].dev_begin = NULL;
                llist[n].dev_end = NULL;
                llist[n].title_begin = NULL;
                llist[n].title_end = NULL;
            }
        } else
        if(section == 2) {
            if(llist[n].dev_begin[0] == ';') {
                // skip comment
                llist[n].dev_begin = NULL;
                llist[n].dev_end = NULL;
                llist[n].title_begin = NULL;
                llist[n].title_end = NULL;
            } else {
                llist[n].boot_item = TRUE;
            }
        } else
        { /* do nothing */; }
    }

    // find lines with same OS, but different parameters
    for(n=0; llist[n].begin; n++) {
        if(!llist[n].boot_item) {
            continue;
        }
        if(llist[n].dupLineN) {
            continue;
        }
        for(m=n+1; llist[m].begin; m++) {
            if(!llist[m].boot_item) {
                continue;
            }
            if(llist[m].dupLineN) {
                continue;
            }
            if(llist[n].dev_len == llist[m].dev_len &&
               !_strnicmp((CCHAR*)(llist[n].dev_begin), (CCHAR*)(llist[m].dev_begin), llist[n].dev_len)) {
                llist[m].dupLineN = n;
            }
        }
    }

    return llist;

} // end parse_boot_ini()

PUCHAR
get_current_arcname()
{
    UNICODE_STRING sys_root;
    UNICODE_STRING sys_root_parsed;
    PUCHAR arcname;
    ULONG Z;
    ULONG T;
    PWCHAR p;

    if(!LocateNTDLLCalls()) {
        return NULL;
    }

#define SYSROOT_PART1_STR   L"\\Device\\Harddisk"
#define SYSROOT_PART1_LEN   ((sizeof(SYSROOT_PART1_STR)-sizeof(WCHAR))/sizeof(WCHAR))

    InitConstUnicodeString(&sys_root, L"\\SystemRoot");
    ParseSymLink1(&sys_root, &sys_root_parsed);

    if(_wcsnicmp(sys_root_parsed.Buffer, SYSROOT_PART1_STR, SYSROOT_PART1_LEN)) {
        return NULL;
    }
    arcname = (PUCHAR)GlobalAlloc(GMEM_FIXED, 128);
    if(1 != swscanf(sys_root_parsed.Buffer+SYSROOT_PART1_LEN,
                    L"%d", &Z)) {
        GlobalFree(arcname);
        return NULL;
    }
    p = wcschr(sys_root_parsed.Buffer+SYSROOT_PART1_LEN, '\\');
    if(!p) {
        GlobalFree(arcname);
        return NULL;
    }

#define SYSROOT_PART2_STR   L"\\Partition"
#define SYSROOT_PART2_LEN   ((sizeof(SYSROOT_PART2_STR)-sizeof(WCHAR))/sizeof(WCHAR))

    if(_wcsnicmp(p, SYSROOT_PART2_STR, SYSROOT_PART2_LEN)) {
        GlobalFree(arcname);
        return NULL;
    }
    if(1 != swscanf(p+SYSROOT_PART2_LEN,
                    L"%d", &T)) {
        GlobalFree(arcname);
        return NULL;
    }
    p = wcschr(p+SYSROOT_PART2_LEN, '\\');
    if(!p) {
        p = L"";
    }

    sprintf((PCHAR)arcname, "multi(0)disk(0)rdisk(%d)partition(%d)%S", Z, T, p);
    return arcname;
} // end get_current_arcname()

ULONG
print_boot_menu(
    PBOOT_INI_LINE_DESCR llist,
    PULONG OrigDefLine,
    PULONG DefLine,
    PULONG CurLine
    )
{
    ULONG i, n;
    ULONG len0;
    ULONG len;
    ULONG len2;
    ULONG len3;
    ULONG line_def0=-1;
    ULONG line_def=-1;
    ULONG line_lst=-1;
    ULONG line_cur=-1;
    UCHAR a, cnt;
    PUCHAR arcname;

    arcname = get_current_arcname();
    if(arcname) {
        len3 = strlen((CCHAR*)arcname);
    } else {
        len3 = 0;
    }

    (*OrigDefLine) = -1;
    (*DefLine) = -1;
    cnt = 0;
    for(n=0; llist[n].begin; n++) {
        if(llist[n].def_item) {
            line_def=n;
        }
        if(llist[n].orig_def_item) {
            line_def0=n;
        }
        if(llist[n].boot_item) {
            if(line_lst == -1) {
                line_lst=n;
            }
            cnt++;
        }
    }
/*    if(line_def0 != -1) {
        line_def = line_def0;
    }*/
    if(line_def == -1) {
        line_def = line_lst;
    }
    if(line_def0 == -1) {
        line_def0 = line_def;
    }
    len = llist[line_def].title_len;
    len0 = llist[line_def0].title_len;
    for(n=line_lst; llist[n].begin; n++) {
        if(!llist[n].boot_item)
            continue;
        len2 = llist[n].dev_len;
        if(len2 == len0) {
            if((*OrigDefLine) == -1 && !_strnicmp((CCHAR*)(llist[n].dev_begin), (CCHAR*)(llist[line_def0].title_begin), len0)) {
                (*OrigDefLine) = n;
                line_def0 = n;
            }
        }
        if(len2 == len) {
            if((*DefLine) == -1 && !_strnicmp((CCHAR*)(llist[n].dev_begin), (CCHAR*)(llist[line_def].title_begin), len)) {
                (*DefLine) = n;
                line_def = n;
            }
        }
        if(len2 == len3) {
            if(line_cur == -1 && !_strnicmp((CCHAR*)(llist[n].dev_begin), (CCHAR*)arcname, len3)) {
                line_cur = n;
            }
        }
    }
    if((*DefLine) == -1) {
        line_def = line_lst;
        (*DefLine) = line_lst;
    }
    if((*OrigDefLine) == -1) {
        line_def0 = line_def;
        (*OrigDefLine) = line_def;
    } else {
        line_def = line_def0;
        (*DefLine) = line_def0;
    }

    set_attr(FOREGROUND_GREEN | FOREGROUND_BLUE | FOREGROUND_RED);

    printf ("Reboot to (%s):\n", BootIniPath);
    printf ("=========\n");

    n = line_def0;
    len2 = llist[n].title_len;
    set_attr(FOREGROUND_GREEN | FOREGROUND_BLUE | FOREGROUND_INTENSITY);
    //printf ("Enter - %.*s [Default]\n", len2, llist[n].title_begin);
    printf ("Enter - Default\n");
    if(line_cur != -1/* && line_cur != n*/) {
        n = line_def0;
        len2 = llist[n].title_len;
        set_attr(FOREGROUND_GREEN | FOREGROUND_INTENSITY);
        //printf ("Space - %.*s [Current]\n", len2, llist[n].title_begin);
        printf ("Space - Current\n");
    }
    set_attr(FOREGROUND_RED | FOREGROUND_INTENSITY);
    //printf ("Enter - %.*s [Default]\n", len2, llist[n].title_begin);
    printf ("Esc   - Cancel\n");
    set_attr(FOREGROUND_GREEN | FOREGROUND_BLUE | FOREGROUND_RED);
    printf ("\n");
    for(i=0, n=line_lst; llist[n].begin; n++) {
        if(!llist[n].boot_item)
            continue;
        len2 = llist[n].title_len;
        if(i < 9) {
            a = '1'+(UCHAR)i;
        } else
        if(i < 9+26) {
            a = 'A'+(UCHAR)i-9;
        } else {
            break;
        }
        if(n == *DefLine) {
            //set_attr(BACKGROUND_GREEN | BACKGROUND_BLUE | BACKGROUND_INTENSITY);
            set_attr(FOREGROUND_GREEN | FOREGROUND_BLUE | FOREGROUND_INTENSITY);
        } else
        if(n == line_cur) {
            set_attr(FOREGROUND_GREEN | FOREGROUND_INTENSITY);
        } else
        {
            set_attr(FOREGROUND_GREEN | FOREGROUND_BLUE | FOREGROUND_RED);
        }
        printf ("%c  -  %.*s ", a, len2, llist[n].title_begin);
        if(n == *DefLine || n == line_cur) {
            if(n == *DefLine) {
                set_attr(FOREGROUND_GREEN | FOREGROUND_BLUE | FOREGROUND_INTENSITY);
                printf("[Default] ");
            }
            if(n == line_cur) {
                set_attr(FOREGROUND_GREEN | FOREGROUND_INTENSITY);
                printf("[Current] ");
            }
        }
        printf("\n");
        i++;
    }
    set_attr(FOREGROUND_GREEN | FOREGROUND_BLUE | FOREGROUND_RED);
    (*CurLine) = line_cur;
    return i;
} // end print_boot_menu()

BOOLEAN
restore_boot_ini(
    char* BootIniPath
    )
{
    HANDLE h;
    ULONG sz;
    PUCHAR buff;
    ULONG read_bytes;
    CHAR BootIniBkpPath[] = "x:\\boot.prv";

    BootIniBkpPath[0] = BootIniPath[0];

    h = CreateFile(BootIniBkpPath,
                      GENERIC_READ, FILE_SHARE_READ, NULL,
                      OPEN_EXISTING,
                      FILE_ATTRIBUTE_NORMAL,
                      NULL);
    if(h == (HANDLE)(-1)) {
        printf("Can't open %s\n", BootIniBkpPath);
        return FALSE;
    }
    sz = GetFileSize(h, NULL);
    buff = (PUCHAR)GlobalAlloc(GMEM_FIXED, sz+1);
    if(!buff) {
        printf("Insufficient memory: can't allocate %d bytes\n", sz);
        return FALSE;
    }
    if(!ReadFile(h, buff, sz, &read_bytes, NULL) ||
       read_bytes != sz) {
        printf("Can't read %s\n", BootIniPath);
        return FALSE;
    }
    CloseHandle(h);
    buff[sz] = 0;

    SetFileAttributes((PCHAR)BootIniPath, FILE_ATTRIBUTE_NORMAL);
    h = CreateFile((PCHAR)BootIniPath, GENERIC_WRITE, FILE_SHARE_READ, NULL,
                      CREATE_ALWAYS,
                      FILE_ATTRIBUTE_NORMAL,
                      NULL);
    if(h == (HANDLE)(-1)) {
        printf("Can't restore %s from backup\n", BootIniPath);
        return FALSE;
    }
    if(!WriteFile(h, buff, sz, &read_bytes, NULL) ||
       read_bytes != sz) {
        printf("Can't write %s\n", BootIniPath);
        return FALSE;
    }
    CloseHandle(h);
    SetFileAttributes((PCHAR)BootIniPath, FILE_ATTRIBUTE_READONLY | FILE_ATTRIBUTE_SYSTEM | FILE_ATTRIBUTE_HIDDEN);

    return TRUE;

} // end restore_boot_ini()

BOOLEAN
build_boot_ini(
    PBOOT_INI_LINE_DESCR llist,
    ULONG LineN,
    ULONG line_def,
    ULONG line_def0,
    char* BootIniPath
    )
{
    ULONG n;
    ULONG cnt;
    ULONG new_line = -1;
    HANDLE h;
    CHAR buff[1024];
    ULONG len;
    ULONG read_bytes;
    LONGLONG ftime1, ftime2, ftime3;

    if((LONG)LineN < 0) {
        new_line = -(LONG)LineN;
    } else {
        cnt = 0;
        for(n=0; llist[n].begin; n++) {
            if(llist[n].boot_item) {
                if(cnt == LineN) {
                    new_line = n;
                    break;
                }
                cnt++;
            }
        }
    }

    ftime1 = 0;
    ftime2 = 0;
    ftime3 = 0;
    h = CreateFile((PCHAR)BootIniPath, GENERIC_READ, FILE_SHARE_READ, NULL,
                      OPEN_EXISTING,
                      FILE_ATTRIBUTE_NORMAL,
                      NULL);
    if(h == (HANDLE)(-1)) {
        printf("Can't open %s\n", BootIniPath);
        return FALSE;
    }
    if(!GetFileTime(h, (PFILETIME)&ftime1, (PFILETIME)&ftime2, (PFILETIME)&ftime3)) {
        ftime1 = 0;
        ftime2 = 0;
        ftime3 = 0;
    }
    CloseHandle(h);

    SetFileAttributes((PCHAR)BootIniPath, FILE_ATTRIBUTE_NORMAL);
    h = CreateFile((PCHAR)BootIniPath, GENERIC_WRITE, FILE_SHARE_READ, NULL,
                      CREATE_ALWAYS,
                      FILE_ATTRIBUTE_NORMAL,
                      NULL);
    if(h == (HANDLE)(-1)) {
        printf("Can't open %s for write\n", BootIniPath);
        restore_boot_ini((PCHAR)BootIniPath);
        return FALSE;
    }

    for(n=0; llist[n].begin; n++) {
        if(llist[n].def_item) {
            len = _snprintf(buff, sizeof(buff), "Default=%.*s\r\n", llist[new_line].dev_len, llist[new_line].dev_begin);
            if(!WriteFile(h, buff, len, &read_bytes, NULL) ||
               read_bytes != len) {
                printf("Can't write %s\n", BootIniPath);
                restore_boot_ini((PCHAR)BootIniPath);
                CloseHandle(h);
                return FALSE;
            }
            len = _snprintf(buff, sizeof(buff), ";OrigDefault=%.*s\r\n", llist[line_def0].dev_len, llist[line_def0].dev_begin);
            if(!WriteFile(h, buff, len, &read_bytes, NULL) ||
               read_bytes != len) {
                printf("Can't write %s\n", BootIniPath);
                CloseHandle(h);
                restore_boot_ini((PCHAR)BootIniPath);
                return FALSE;
            }
        } else
        if(llist[n].orig_def_item) {
            // do nothing
            continue;
        } else {
            if(llist[new_line].dupLineN && llist[n].LineN == llist[new_line].dupLineN) {
                len = llist[new_line].len;
                if(!WriteFile(h, llist[new_line].begin, len, &read_bytes, NULL) ||
                   read_bytes != len) {
                    printf("Can't write %s\n", BootIniPath);
                    CloseHandle(h);
                    restore_boot_ini((PCHAR)BootIniPath);
                    return FALSE;
                }
            }
            if(llist[n].LineN == new_line && llist[new_line].dupLineN) {
                // do nothing
                continue;
            } else {
                len = llist[n].len;
                if(!WriteFile(h, llist[n].begin, len, &read_bytes, NULL) ||
                   read_bytes != len) {
                    printf("Can't write %s\n", BootIniPath);
                    CloseHandle(h);
                    restore_boot_ini((PCHAR)BootIniPath);
                    return FALSE;
                }
            }
        }
    }

    SetFileAttributes((PCHAR)BootIniPath, FILE_ATTRIBUTE_READONLY | FILE_ATTRIBUTE_SYSTEM | FILE_ATTRIBUTE_HIDDEN);
    if(ftime1) {
        // restore file times
        if(!SetFileTime(h, (PFILETIME)&ftime1, (PFILETIME)&ftime2, (PFILETIME)&ftime3)) {
            printf("Can't update times for %s\n", BootIniPath);
        }
    }

    CloseHandle(h);

    return TRUE;
} // end build_boot_ini()

BOOLEAN
FindBootIni(
    HWND hDlgItem
    )
{
    UCHAR a, d;
    ULONG cnt;
    HKEY hKey;
    ULONG flags;
    ULONG tmp;
    CHAR Drive[] = "x:\\";

    cnt = 0;
    for(a='A'; a<='Z'; a++) {
        Drive[0] = a;
        if(GetDriveType(Drive) != DRIVE_FIXED)
            continue;
        BootIniPath[0] = a;
        flags = GetFileAttributes((PCHAR)&(BootIniPath[0]));
        if(flags != 0xffffffff) {
            if(!cnt) {
                printf("Found BOOT.INI at\n");
                d = a;
                //break;
            }
            cnt++;
            printf("    %s\n", BootIniPath);
            if(hDlgItem) {
                SendMessage(hDlgItem, CB_ADDSTRING, 0, (LPARAM)BootIniPath);
            }
        }
    }
    if(!cnt) {
        printf("No BOOT.INI found\n");
        return FALSE;
    }

    BootDrv[0] = d;
    BootDrv[1] = ':';
    BootDrv[2] = 0;

    tmp = SetEnvironmentVariable(STR_BOOT_DRV_ENV_VAR, (PCHAR)BootDrv);
    if(!tmp) {
        printf("Can't setup SystemDrive environment variable\n");
    }
    tmp = RegOpenKeyEx(HKEY_LOCAL_MACHINE, REG_PATH_ENV, 0, KEY_ALL_ACCESS, &hKey);
    if(tmp != ERROR_SUCCESS) {
        printf("Can't access system environment variables\n");
    } else {
        tmp = RegSetValueEx(hKey, STR_BOOT_DRV_ENV_VAR, NULL, REG_SZ, BootDrv, 3);
        if(tmp != ERROR_SUCCESS) {
            printf("Can't setup SystemDrive system environment variable\n");
        }
        RegCloseKey(hKey);
    }

    BootIniPath[0] = BootDrv[0];
    return TRUE;

} // end FindBootIni()

ULONG
__stdcall
RebootImmediately(PVOID);

//void main(int argc, char** argv)
ULONG
__stdcall
RebootManager(PVOID)
{
    ULONG len;
    UCHAR a, d;
    ULONG flags;
    ULONG cnt;
    ULONG key = 0;
    INPUT_RECORD inp;
    HKEY hKey;
    ULONG tmp;
    PBOOT_INI_LINE_DESCR llist;
    ULONG cur_item;
    ULONG def_item;
    ULONG def0_item;
    ULONG n;
    CHAR Drive[] = "x:\\";

    AllocConsole();
    StdIn  = GetStdHandle(STD_INPUT_HANDLE);
    if(StdIn == INVALID_HANDLE_VALUE || !StdIn) {
        AllocConsole();
        StdIn  = GetStdHandle(STD_INPUT_HANDLE);
        if(StdIn == INVALID_HANDLE_VALUE || !StdIn) {
            FreeConsole();
            MessageBox(NULL, "Can't init console\n", "Error", MB_OK | MB_ICONERROR);
            RebootInitiated = FALSE;
            return -1;
        }
        OwnConsole = TRUE;
    }
    StdOut = GetStdHandle(STD_OUTPUT_HANDLE);
    SetConsoleTitle("Reboot manager");

    len = strlen((CCHAR*)BootDrv);
/*
    tmp = RegOpenKeyEx(HKEY_LOCAL_MACHINE, REG_PATH_ENV, 0, KEY_ALL_ACCESS, &hKey);
    if(tmp != ERROR_SUCCESS) {
        //printf("Can't access system environment variables\n");
    } else {
        n = sizeof(BootDrv);
        tmp = RegQueryValueEx(hKey, STR_BOOT_DRV_ENV_VAR, NULL, NULL, BootDrv, &n);
        if(tmp != ERROR_SUCCESS) {
            //printf("Can't setup SystemDrive system environment variable\n");
        } else {
        }
        RegCloseKey(hKey);
        len = n;
    }

    if(!len) {
        len = GetEnvironmentVariable(STR_BOOT_DRV_ENV_VAR, (PCHAR)BootDrv, sizeof(BootDrv));
    }
*/
    if(!len) {
        cnt = 0;
        for(a='c'; a<='z'; a++) {
            Drive[0] = a;
            if(GetDriveType(Drive) != DRIVE_FIXED)
                continue;
            BootIniPath[0] = a;
            flags = GetFileAttributes((PCHAR)&(BootIniPath[0]));
            if(flags != 0xffffffff) {
                if(!cnt) {
                    printf("Found BOOT.INI at\n");
                    d = a;
                }
                cnt++;
                printf("    %s\n", BootIniPath);
            }
        }
        if(!cnt) {
            printf("No BOOT.INI found\n");
            RebootInitiated = FALSE;
            FreeConsole();
            return -1;
        }
        if(cnt > 1) {
            printf("please hit proper key (A-Z) to chose proper one\n");
            while(TRUE) {
                key = ReadConsoleInput(StdIn, &inp, 1, &tmp);
                if(key && inp.EventType == KEY_EVENT && inp.Event.KeyEvent.bKeyDown) {
                    if(inp.Event.KeyEvent.wVirtualScanCode == 0x01 /*SCAN_ESC*/) {
                        RebootInitiated = FALSE;
                        FreeConsole();
                        return (1);
                    }
                    a = inp.Event.KeyEvent.uChar.AsciiChar;
                    tmp = drv_letter_to_index(a);
                    if(tmp != -1) {
                        Drive[0] = a;
                        if(GetDriveType(Drive) != DRIVE_FIXED)
                            continue;
                        BootIniPath[0] = a;
                        flags = GetFileAttributes((CCHAR*)BootIniPath);
                        if(flags != 0xffffffff) {
                            d = (UCHAR)tmp + 'a';
                            BootIniPath[0] = d;
                            printf("Using %s\n", BootIniPath);
                            break;
                        }
                    }

                }
                Sleep(50);
            }
        } else {
            BootIniPath[0] = d;
        }
        BootDrv[0] = d;
        BootDrv[1] = ':';
        BootDrv[2] = 0;
        tmp = SetEnvironmentVariable(STR_BOOT_DRV_ENV_VAR, (PCHAR)BootDrv);
        if(!tmp) {
            printf("Can't setup SystemDrive environment variable\n");
        }
        tmp = RegOpenKeyEx(HKEY_LOCAL_MACHINE, REG_PATH_ENV, 0, KEY_ALL_ACCESS, &hKey);
        if(tmp != ERROR_SUCCESS) {
            printf("Can't access system environment variables\n");
        } else {
            tmp = RegSetValueEx(hKey, STR_BOOT_DRV_ENV_VAR, NULL, REG_SZ, BootDrv, 3);
            if(tmp != ERROR_SUCCESS) {
                printf("Can't setup SystemDrive system environment variable\n");
            }
            RegCloseKey(hKey);
        }
    }
    BootIniPath[0] = BootDrv[0];

    llist = parse_boot_ini((PCHAR)BootIniPath);
    if(!llist) {
        if(OwnConsole) {
            MessageBox(NULL, "Can't understand BOOT.INI format\n", "Error", MB_OK | MB_ICONERROR);
        }
        FreeConsole();
        return -1;
    }

    n = print_boot_menu(llist, &def0_item, &def_item, &cur_item);

    while(TRUE) {
        key = ReadConsoleInput(StdIn, &inp, 1, &tmp);
        if(key && inp.EventType == KEY_EVENT && inp.Event.KeyEvent.bKeyDown) {
            if(inp.Event.KeyEvent.wVirtualScanCode == 0x01 /*SC_ESC*/) {
                RebootInitiated = FALSE;
                FreeConsole();
                return (1);
            }
            a = inp.Event.KeyEvent.uChar.AsciiChar;
            if(a == ' ') {
                tmp = -(LONG)(llist[cur_item].LineN);
                break;
            }
            if(inp.Event.KeyEvent.wVirtualScanCode == 0x1c /*SC_ENTER*/) {
                tmp = -(LONG)(llist[def0_item].LineN);
                break;
            }
            tmp = num_or_letter_to_index(a);
            if(tmp != -1) {
                break;
            }

        }
        Sleep(50);
    }

    if(!build_boot_ini(llist, tmp, def_item, def0_item, (PCHAR)BootIniPath)) {
        if(OwnConsole) {
            MessageBox(NULL, "Can't construct new BOOT.INI\n", "Error", MB_OK | MB_ICONERROR);
        }
        FreeConsole();
        return -1;
    }
    FreeConsole();

    RebootImmediately(NULL);

    RebootInitiated = FALSE;
    //exit (0);
    return 0;
} // end RebootManager()
